/**
 * server - file server in UDP protocol.
 * 2000-2001 by AUTHORS (see AUTHORS for more informations).
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License.  See the file COPYING in the main directory of the
 * gnu/poker distribution for more details.
 *
 **************************************************************************
 * get.c :  allow to get a file.
 ************************************************************************** 
 * dim jan 20 15:39:48 CET 2002 - clucas - first edition
 * dim jan 20 21:45:54 CET 2002 - clucas - stable implementation
 **************************************************************************
 */

#include "common.h"
#include "ftp.h"
#include <time.h>

extern int errno ;

/**
 * server-receive:
 * The server receive a file from a remote host.
 * parameter(s):
 * o fd		: descriptor of socket
 * o s		: sockadd
 * o p_low	: packet from lowserver
 * return:
 * o	int
 */
int server_receive( int fd , struct sockaddr_in *s , packet *p_low)
{
	packet		p;
	int		retval , nb_chunk , nb_packet , lg;
	int		tmp_chunk , tmp_packet , fd_file ;
	int 		chunk_length , num , current_chunk , donnees , i, transfert_state ;
	data_block	*tab ;
	time_t		h1 , h2;
	char		file[NB_PACKET_TO_SEND];

	
	/* initialisation de certaines variables */
	lg = sizeof( struct sockaddr_in );
	donnees = i = chunk_length = 0;

	strncpy( file , p_low->type_packet.control.liste , NB_PACKET_TO_SEND );

	/* opening the file we want to download */
	fd_file = open( file , O_CREAT|O_WRONLY , 0666 );
	if( fd_file == -1 ) {
		if( errno == EACCES ){
			p.header = CTRL_PACKET ;
			p.type_packet.control.ordre = PERMISSION_DENIED ;
			sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) ); 
		} else {
			fprintf(stderr,"lowserver> get> open failed !\n");
			return -1 ;
		}
	}

	/* code */
	p.header = CTRL_PACKET ;
	p.type_packet.control.ordre = ACK_SEND ;
	sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) ); 

	/* we are waiting for the nb_chunk */
	retval = my_read( fd , &p , sizeof(p) , (struct sockaddr *)s , &lg ) ;
	if( retval < 0 ) {
		fprintf(stderr,"lowserver> get> my_read\n");
		return -1;
	} else { 	
		if( p.type_packet.control.ordre == BEGIN_GET ) {
			nb_chunk = p.type_packet.control.gen_num ;
			p.header = CTRL_PACKET ;
			p.type_packet.control.ordre = ACK_BEGIN_GET ;
			sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) );
		} else {
			printf("lowserver> server-receive> Wrong message.\n");
			return -1;
		}
	}
	
	/* we are waiting for the nb_pakcet */
	retval = my_read( fd , &p , sizeof(p) , (struct sockaddr *)s , &lg ) ;
	if( retval < 0 ) {
		fprintf(stderr,"lowserver> get> my_read\n");
		return -1;
	} else {
		if( p.type_packet.control.ordre == BEGIN_DATA ) {
			nb_packet = p.type_packet.control.gen_num ;
			p.header = CTRL_PACKET ;
			p.type_packet.control.ordre = ACK_BEGIN_DATA ;
			sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) );
		} else {
			printf("lowserver> server-receive> Wrong message.\n");
			return -1;
		}
	}

	tab = create_tab( );		
	
/* 
 * FIXME:
 * This part of code is awfull: i know !!
 * If i have more time i'll code this in a better way !
 */

	while( donnees == 0  ) {
		bzero( &p , sizeof(packet) );
		retval = titux_read( fd , &p , sizeof(p) , (struct sockaddr *)s , &lg ) ;
		switch( retval ){
			case CTRL_PACKET: {
				switch( p.type_packet.control.ordre ) {
					case END_CHUNK : {
						if( chunk_length == i ) {
							i = 0 ;
							/* all packets of this chunk are received */
							if( fflush_tab( fd_file , tab , chunk_length ) != 0 ) {
								p.header = CTRL_PACKET ; 
								p.type_packet.control.ordre = ERR_NO_SPACE ;
								sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,
												sizeof(struct sockaddr) ); 	
								close( fd_file );
								unlink( file );	
								return -1 ;
							}
							/* we init tab for further chunk */
							init_data_tab( tab );
							/* we send a CHUNK_OK message */
							bzero( &p , sizeof(packet) );
							p.header = CTRL_PACKET ; 
							p.type_packet.control.ordre = CHUNK_OK ;
							sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) ); 

							if( current_chunk == nb_chunk )
								transfert_state = 1 ;

						} else {
							/* some packets have bee losen */
							if( i < chunk_length ) {
								p=update_ctrl_msg_list( tab , chunk_length);
								p.header = CTRL_PACKET ;
								sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,
											sizeof(struct sockaddr) ); 
							} else
								i = chunk_length ;
						}		
					}
					break;
			
					case END_GET: {
						/* this is the end ! ;) */
						donnees = 1 ;
						if( transfert_state != 1 ) {
							/* 
							 * we reached the end of transfert without all data:
							 * OUCH !!!! ;-(
							 */
							fprintf(stderr,"lowserver> get> Error in transfert !\n");
							close( fd_file );
							return -1;
						} else {
							unsigned int 		taille ;
							taille       =  p.type_packet.control.gen_num ;
							time( &h2 );
							printf("%d bytes send (%lds).",taille,difftime( h2 , h1 )) ;

						}
					}
					break;	
					default: fprintf(stderr,"lowserver> get> ctrl_read> I don't"
							 "understand what he said!(%d)\n",p.type_packet.control.ordre);
				} /* switch */
			}
			break ;
			case DATA_PACKET: {
				switch( p.type_packet.donnees.ordre ) {
					case BEGIN_CHUNK: {
						int size ;
						num = p.type_packet.donnees.num_packet ;
						chunk_length = p.type_packet.donnees.chunk_length ;
						current_chunk = p.type_packet.donnees.num_chunk ;
						tab[ num ].length = size = p.type_packet.donnees.packet_size ;
						tab[ num ].ok = 1 ;
						memcpy( tab[ num ].data , p.type_packet.donnees.data , size ); 
						/* i++: sert uniquement pour savoir le nb de paquets recus. */
						i++;
					}
					break;

					default:
						fprintf(stderr,"lowserver> get> Problem: message ununderstood !\n");
						return -1 ;
				
				}/* switch */	
			}
			break;
			case -1 : {
				fprintf(stderr,"lowserver> get> titux_read => Error with the server !( mesg is not data or control packet)\n");
				return -1; 
			}
			break;
			default:
		}/* switch */
	}/*while*/


	close( fd_file );
	free( tab );
return 0;
}

/**
 * server_send:
 * this is the function which must be executed by the server
 * to answer to a get request (see below).
 * parameter(s):
 * o fd		: socket descriptor
 * o s		: sockaddr
 * o file	: filename
 * o dir	: where i send the file
 * return:
 * int 
 */
int client_send( int fd , struct sockaddr_in *s , char *file , char *dir)
{
	packet		p;
	int		retval, lg , size, nb_chunk, nb_packet;
	int		fd_file, tmp_chunk , tmp_packet; 
	int		cpt , nb , current_chunk ;
	int		cpt_packet , cpt_packet_loosen ;
	data_block	*tab=NULL;
	char		f_file[255];
	char		f_dir[255];
	time_t		h1 , h2 ;

	h1 = time( NULL );

	strncpy( f_file , file , 255 );
	strncpy( f_dir , dir , 255 );

	lg = sizeof( struct sockaddr );
	bzero( &p , sizeof(packet) );

	/* we verify some things on the file asked */
	retval = test_file( f_file );
	if( retval > 0 ) {
		size = retval ;
		p.header = CTRL_PACKET ;
		p.type_packet.control.ordre = SEND ;
		strncpy( p.type_packet.control.liste , f_file , NB_PACKET_TO_SEND );
		p.type_packet.control.gen_num = size ;
		if( (retval = resend( fd , &p , sizeof(packet) , ACK_SEND , (struct sockaddr *)s , lg )) < 0 ) {
			switch( retval ) {
				case PERMISSION_DENIED:
					printf("You can not send this file in %s.\n", f_dir);
					break;

				case FILE_TOO_LARGE :
					/* NOT IMPLEMENTED NOW => Cf quotactl */
					printf("Not enough space on remote server disk.\n");	
					break;
				default: printf("Wrong message !!!!\n");
			}/* switch */
			return retval ;
		}
	} else {
		if( retval == -1 ) {
			/* le fichier n'existe pas ! */
			printf("source file does not exist.\n") ;
			return FILE_NOT_FOUND ;
		} else {
			switch( retval ) {
				case -ERR_FILE_IS_DIR :{
					printf("source file is a directory, i can not send this.\n");
				}
				break;
		
				case -ERR_FILE_IS_LINK :{
					printf("source file is a symbolic link, i can not send this.\n");
				}
				break;

				case -ERR_FILE_IS_SOCKET : {
					printf("source file is a socket, i can not send this.\n");
				}
				break;
			}
			return retval ;
		}
	}

	/* we calculate 'nb_packet' */
	nb_packet =( (int)size % DATA_SIZE == 0 ? (int)size/DATA_SIZE : (int)size/DATA_SIZE + 1 ) ;
	/* after this: 'nb_chunck' */
	nb_chunk = ( (int)nb_packet % PACKET_PER_CHUNK == 0 ? (int)nb_packet/PACKET_PER_CHUNK : (int)nb_packet/PACKET_PER_CHUNK + 1 );	
	
	/* we send a BEGIN_GET packet */
	bzero( &p , sizeof(packet) );
	p.header = CTRL_PACKET ;
	p.type_packet.control.ordre = BEGIN_GET ;
	p.type_packet.control.gen_num = nb_chunk ;
	if( resend( fd , &p , sizeof(packet) , ACK_BEGIN_GET , (struct sockaddr *)s , sizeof( struct sockaddr) ) < 0 ) {
		close( fd );
		kill_lowserver( );
	}

	/* we send a BEGIN_GET packet */
	bzero( &p , sizeof(packet) );
	p.header = CTRL_PACKET ;
	p.type_packet.control.ordre = BEGIN_DATA ;
	p.type_packet.control.gen_num = nb_packet ;
	if( resend( fd , &p , sizeof(packet) , ACK_BEGIN_DATA , (struct sockaddr *)s , sizeof( struct sockaddr) ) < 0 ) {
		close( fd );
		kill_lowserver( );
	}

	fd_file = open( f_file , O_RDONLY, 0666 );
	if( fd_file == -1 ) {
		fprintf(stderr,"lowserver> server_get> open failed !\n");
		return -1 ;
	}		
	
	tab = create_tab( );

	tmp_chunk = tmp_packet = current_chunk = cpt_packet = cpt_packet_loosen = 0;
	while( tmp_chunk < nb_chunk ) {
		/* -- variables locales */
		int 		i,j;

		/* if the previous chunk is completely send, we read some data */
		if( current_chunk == 0 ) 
			nb = fill_data_tab( tab , fd_file );

		bzero( &p , sizeof(packet) );
		p.header = DATA_PACKET ;
		p.type_packet.donnees.chunk_length = nb ;
		p.type_packet.donnees.ordre = BEGIN_CHUNK ;
		p.type_packet.donnees.num_chunk = tmp_chunk ;

		j=0;
		for( i=0 ; i<nb ; i++ ) {
			if( tab[i].ok == -1 ) {
				p.type_packet.donnees.num_packet = i ;
				p.type_packet.donnees.packet_size = tab[i].length ;
				memcpy( p.type_packet.donnees.data , tab[i].data , tab[i].length ) ;
				/* finally: we send the packet */
				sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) ); 
				/* we have send this packet */
				tab[i].ok = 1 ;
				cpt_packet ++ ;
				j++ ; 
			}
		}

		/* on force la boucle du resend */
		retval = -1 ;	
		while( retval == -1 ) {
		/* afterwards: we send a 'END_CHUNK' message */
		bzero( &p , sizeof(packet) );
		p.header = CTRL_PACKET ;
		p.type_packet.control.ordre = END_CHUNK;  
				if( (retval = resend( fd , &p , sizeof(packet) , CHUNK_OK , (struct sockaddr *)s , sizeof(struct sockaddr) )) <= 0 ) {
				switch( retval ) {
					case -CHUNK_NOT_OK: {
						int tmp ,k ;
						tmp = 0 ;
						/* we say we resay some packets in this chunk */
						current_chunk = 1 ;
						
						/* we update array of data_block */
						tmp = update_data_block( tab , p.type_packet.control.liste , nb );	
						cpt_packet_loosen += tmp ;
						}
					break;
					case 0:
						if( tmp_chunk == nb_chunk-1 ) {
							bzero( &p , sizeof(packet) );
							p.header = CTRL_PACKET ;
							p.type_packet.control.ordre = END_GET ;
							p.type_packet.control.gen_num  = size ;
							sendto(fd, &p, sizeof(packet),0, (struct sockaddr *)s, 
											sizeof(struct sockaddr) ); 
						} else {
							/* flag changed to rebuild array of data_block */
							current_chunk = 0 ;
						}				
			
						/* we incremente the chunk counter :) */
						tmp_chunk ++ ;
					break;
					case ERR_NO_SPACE:{
						fprintf(stderr,"Not enough space on remote host.\n");
						close( fd_file );
						return -1 ;
					}
					break;
					default: /* on force la boucle  continuer */
						retval = -1 ;
				}/* switch */
			}/* if */ 
		}/* while */
	} /* while tmp_chunk */	

	/* we closed file descriptor */
	close( fd_file );
	/* we free tab */
	free( tab );
	
	h2 = time( NULL );

printf("send packets:%d, loosen packets:%d (%.2lf %%) in %.2lf s\n",cpt_packet,
		(float)cpt_packet_loosen,cpt_packet_loosen/cpt_packet, difftime( h2 , h1 ));
return 0;
}
