/**
 * server - file server in UDP protocol.
 * 2000-2001 by AUTHORS (see AUTHORS for more informations).
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License.  See the file COPYING in the main directory of the
 * gnu/poker distribution for more details.
 *
 **************************************************************************
 * utils.c :  some functions which are useful in get and send functions.
 * 
 ************************************************************************** 
 * dim jan 20 21:55:23 CET 2002 - clucas - first edition
 **************************************************************************
 */

#include "common.h"
#include "ftp.h"

/**
 * fill_data_tab:
 * this function fill the tab given by reference with data 
 * from the file given too to the function
 * so...
 * parameter(s):
 * o tab : array of data
 * o fd : file descriptor
 * return:
 * int : nb_packet read
 */
int fill_data_tab( data_block *tab , int fd )
{
	int		nb, retval ,ok;
	char		buffer[DATA_SIZE];

	nb = ok = 0 ;
	while( (ok ==0 ) && (retval = read( fd , buffer , sizeof(buffer)) ) > 0 ) {
		memcpy( tab[nb].data , buffer , sizeof(buffer) );
		tab[nb].ok = -1 ;
		tab[nb].length = retval ;
		nb ++; 
		if( nb > PACKET_PER_CHUNK - 1 ) 
			ok = 1 ;
		else	ok = 0 ;
	} 

	if( retval == -1 ) {
		fprintf(stderr,"lowserver> fill_data_tab> read !\n");
		return -1 ;
	}

return nb ;
}


/**
 * init_data_tab:
 * this function init the tab given by reference
 * this function is used in get();
 * so...
 * parameter(s):
 * o tab : array of data
 * return:
 * void function
 */
void init_data_tab( data_block *tab )
{
	register int		i;
	for( i=0 ; i<PACKET_PER_CHUNK ; i++) {
		tab[i].ok = -1 ;
		tab[i].length = 0 ;
	}
}


/*
 * update_data_block:
 * this function has for goal to upadate blocks, which have been received
 * by the client ( by means of the 'ok' variable).
 * parameters:
 * o	t 	= data_bloc array
 * o	list 	= list of block which must be resend
 * o	nb_block= number of block, which is 'chunk'. 
 * return:
 * int : nb packet loosen.
 */
int update_data_block( data_block *t , char *list , int nb_block )
{
	register int 	i ;
	int		j ;

	j = 0;
	for( i=0 ; i<nb_block ; i++ ) {
		/* packet i has been received */
		if( list[i] == 0 ) {
			t[i].ok = 1 ;
		} else {
			t[i].ok = -1 ;
			j++;
		}
	}
return j;
}

/**
 * update_ctrl_msg_list:
 * this function has for goal to update a list of packets
 * which are not been received.
 * So that we ask to the server to resend these packets.
 * parameters:
 *
 * return:
 */
packet update_ctrl_msg_list( data_block *t , int nb_block )
{
	register int		i,j;
	packet			temp ;

	bzero( &temp , sizeof(packet) );
	temp.header = CTRL_PACKET ; 
	temp.type_packet.control.ordre = CHUNK_NOT_OK ;

	j=0;
	for( i=0 ; i<nb_block ; i++ ) {
		if( t[i].ok == 1 ) {
			temp.type_packet.control.liste[i] = 0 ;
			j ++ ;
		} else
			temp.type_packet.control.liste[i] = 1 ;
	}
return temp ;
}


/**
 * fflush_tab:
 * 
 * parameter(s):
 * o fd_file:
 * o tab : 
 * o num_packet:
 * return:
 * this a void function
 */
int fflush_tab( int fd_file , data_block *tab , int num_packet )
{
	register int 		i;
	int			ret ;
	
	for( i=0 ; i<num_packet ; i++ ){
		ret = write( fd_file , tab[i].data , tab[i].length );
		if( ret == -1 ) {
			if( errno == ENOSPC ) {
				return ERR_NO_SPACE ;
			}
		}
	}
return 0;
}

/**
 *
 *
 */
int test_file( char *file )
{
	extern int		errno ;
	int			retval , ret ;
	struct stat		st_file ;
	
	retval = stat( file , &st_file );
	if( retval == -1 ) {
		if( errno == ENOENT ) {
			return -1 ;
		} else {
			printf("problme avec le fichier %s!!\n",file);
		}

	} else {
		ret = S_ISDIR( st_file.st_mode );
		if( ret == 1 )
			return -ERR_FILE_IS_DIR ;
		
		ret = S_ISLNK( st_file.st_mode );
		if( ret == 1 )
			return -ERR_FILE_IS_LINK ;

		ret = S_ISSOCK( st_file.st_mode );
		if( ret == 1 )
			return -ERR_FILE_IS_SOCKET ;

		/* this is a correct file for transfert */
		return st_file.st_size ; 
	}
}


/**
 * create_tab:
 * this function create data array (static array).
 * parameter(s):
 * o void 
 * return:
 * o pointer on this array
 */
data_block* create_tab( void )
{
	data_block		*tmp = NULL ;
	int			i , taille_champ ;	

	tmp = (data_block *)malloc( PACKET_PER_CHUNK*sizeof( data_block ) );
	if( tmp == NULL ) {
		fprintf( stderr , "lowserver> create_tab> malloc !\n");
		exit(1);
	}

	for( i=0 ; i<PACKET_PER_CHUNK ; i++ ){
		tmp[i].ok = -1 ;
	}

	return tmp ;
}
