/**
 * server - file server in UDP protocol.
 * 2000-2001 by AUTHORS (see AUTHORS for more informations).
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License.  See the file COPYING in the main directory of the
 * gnu/poker distribution for more details.
 *
 * ls.c :  function for the 'ls'.
 * 
 * dim nov 25 11:54:58 CET 2001 - clucas - first edition
 * Sun Dec  2 21:52:22 CET 2001 - clucas - pre_ls implementation
 * mar dc 11 10:23:50 CET 2001 - clucas - pre_ls implementation with bloc_ls
 * mar dc 11 16:00:01 CET 2001 - clucas - bug fix on realloc in pre_ls
 * jeu jan 17 21:37:31 CET 2002 - clucas - implementation of ls2
 * ven jan 18 22:41:20 CET 2002 - clucas - bug fix on send_data_ls (close(fd_file);)
 */
#include "common.h"
#include "server.h"
#include "../lib/ftp.h"


/**
 * ls2:
 * This is a second implementation of the 'remote ls'.
 * parameter(s):
 * o fd		: socket descriptor
 * o s		: structure which contain address on remote host
 * return:
 * this is a void function
 *
 */
void ls2( int fd , struct sockaddr_in *s)
{
	int		pid , fd_file  ;
	char		home[255];
	
	strncpy( home , getenv( "HOME" ) , 255 );
	strncat( home , "/ls.tmp", 255 );

	fd_file = open( home , O_WRONLY|O_CREAT|O_TRUNC , 0666 );
	if( fd_file < 0) {
		perror("lowserver>> ls2> open ");
		exit(1);
	}

	switch( pid = fork() ) {
		case -1: perror("lowserver>> fork");
			 exit(1);
		case 0: close( 1 );
			dup( fd_file );
			close( fd_file );
			execlp( "ls" , "ls" , "-lh" , NULL );
			fprintf(stderr, "lowserver>> exec in child has failed !\n");
			exit(1);
			break;

		default: wait(NULL);
			 close( fd_file );
	}

	/* we call the function which send ls_data */
	send_data_ls( fd , s , home );

	unlink( home );
}

/**
 * send_data_ls:
 *
 * parameter(s):
 * o fd 	: socket descriptor
 * o s		: address of the remote host
 * o home	: where i must find the 'ls.tmp'.
 * return:
 * int  : status of the send
 */
int send_data_ls( int fd , struct sockaddr_in *s , const char *home )
{
	packet		p;
	int		retval, lg , nb_chunk , nb_packet ;
	int		fd_file, tmp_chunk , tmp_packet; 
	int		cpt , nb , current_chunk ;
	int		cpt_packet , cpt_packet_loosen ;
	long		size ;
	data_block	*tab=NULL;
	struct stat	st_file ;

/*
 * current_chunk :
 * 	=> 1 : if we are sending or resending some packets
 * 	=> 0 : if we must build and send a new chunk if file
 *
 *
 */
	lg = sizeof( struct sockaddr );

	/* now we watch for the size of the 'ls.tmp' file */
	retval = stat( home , &st_file );
	if( retval == -1 ){
		if( errno == ENOENT ) {
			return -1 ;
		} else {
			printf("problme avec le path!\n");
		}
	} else { size = st_file.st_size ; }

	/* we calculate 'nb_packet' */
	nb_packet =( (int)size % DATA_SIZE == 0 ? (int)size/DATA_SIZE : (int)size/DATA_SIZE + 1 ) ;
	/* after this: 'nb_chunck' */
	nb_chunk = ( (int)nb_packet % PACKET_PER_CHUNK == 0 ? (int)nb_packet/PACKET_PER_CHUNK : (int)nb_packet/PACKET_PER_CHUNK + 1 );	
	
	/* 
	 * First of all:
	 * we respond to the remote host:
	 * 'this is ok for your ls! ' 
	 */
	bzero( &p , sizeof(packet) );	
	p.header = CTRL_PACKET ;
	p.type_packet.control.ordre = ACK_LS ;
	p.type_packet.control.gen_num = nb_chunk ;
	sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) ); 

	/* pour vrifier qq chose */
	oups: 
	retval = -1 ;
	while( retval == -1 ) {
		/* we send a BEGIN_DATA packet */
		bzero( &p , sizeof(packet) );	
		p.header = CTRL_PACKET ;
		p.type_packet.control.ordre = BEGIN_DATA ;
		p.type_packet.control.gen_num = nb_packet ;
		if( (retval=resend( fd , &p , sizeof(packet) , ACK_BEGIN_DATA , (struct sockaddr *)s , sizeof( struct sockaddr) )) < 0 ) {
			switch( retval ) {
				case -LS: {
					bzero( &p , sizeof(packet) );	
					p.header = CTRL_PACKET ;
					p.type_packet.control.ordre = ACK_LS ;
					p.type_packet.control.gen_num = nb_chunk ;
					sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) ); 
	
					retval = -1;
				}
				break;
				default: printf("resend sur ACK_BEGIN_dATA:retval=%d\n",retval);close( fd ); kill_lowserver( );
			}
		}/* if */
	}/* while */
	
	fd_file = open( home , O_RDONLY, 0666 );
	if( fd_file == -1 ) {
		fprintf(stderr,"lowserver> ls> send_data_ls> open failed !\n");
		return -1 ;
	}		
	
	tab = create_tab( );

	tmp_chunk = tmp_packet = current_chunk = cpt_packet = cpt_packet_loosen = 0;
	while( tmp_chunk < nb_chunk ) {
		/* -- variables locales */
		int 		i,j;

		/* if the previous chunk is completely send, we read some data */
		if( current_chunk == 0 ) 
			nb = fill_data_tab( tab , fd_file );

		bzero( &p , sizeof(packet) );
		p.header = DATA_PACKET ;
		p.type_packet.donnees.chunk_length = nb ;
		p.type_packet.donnees.ordre = BEGIN_CHUNK ;
		p.type_packet.donnees.num_chunk = tmp_chunk ;

		j=0;
		for( i=0 ; i<nb ; i++ ) {
			if( tab[i].ok == -1 ) {
				p.type_packet.donnees.num_packet = i ;
				p.type_packet.donnees.packet_size = tab[i].length ;
				memcpy( p.type_packet.donnees.data , tab[i].data , tab[i].length ) ;
				/* finally: we send the packet */
				sendto( fd , &p , sizeof(packet),0,(struct sockaddr *)s,sizeof(struct sockaddr) ); 
				/* usleep(5); */
				/* we have send this packet */
				tab[i].ok = 1 ;
				cpt_packet ++ ;
				j++ ; 
			}
		}
	
		/* afterwards: we send a 'END_CHUNK' message */
		p.header = CTRL_PACKET ;
		p.type_packet.control.ordre = END_CHUNK;  
		if( (retval = resend( fd , &p , sizeof(packet) , CHUNK_OK , (struct sockaddr *)s , sizeof(struct sockaddr) )) <= 0 ) {
			switch( retval ) {
				case -CHUNK_NOT_OK: {
					int tmp ,k ;
					tmp = 0 ;
					/* we say we resay some packets in this chunk */
					current_chunk = 1 ;
					
					/* we update array of data_block */
					tmp = update_data_block( tab , p.type_packet.control.liste , nb );	
					cpt_packet_loosen += tmp ;
					}
				break;
				case 0:
					if( tmp_chunk == nb_chunk-1 ) {
						p.type_packet.control.ordre = END_GET ;
						p.type_packet.control.gen_num  = size ;
						sendto(fd, &p, sizeof(packet),0, (struct sockaddr *)s, sizeof(struct sockaddr) ); 
					} else {
						/* flag changed to rebuild array of data_block */
						current_chunk = 0 ;
					}				
		
					/* we incremente the chunk counter :) */
					tmp_chunk ++ ;
				break;
				default:
					return retval;
			}/* switch */
		}/* if */ 
	} /* while tmp_chunk */	

	/* we closed file descriptor */
	close( fd_file );
	/* we free tab */
	free( tab );
return 0;
}

/*
 * update_ls_block:
 * this function have for goal to upadate blocks, which have been received
 * by the client ( by means of the 'ok' variable).
 * parameters:
 * o	t 	= ls_bloc array
 * o	list 	= list of block which must be resend
 * o	nb_block= number of block, which is 'ls'. 
 * return:
 * this is a void-function.
 */
void update_ls_block( bloc_ls *t , char *list , int nb_block )
{
	int i ;

	for( i=0 ; i<nb_block ; i++ ) {
		/* packet i has been received */
		if( list[i] == 0 ) {
			t[i].ok = 1 ;
		} else {
			t[i].ok = -1 ;
		}
	}
}

/*
 * send_ls:
 * This function send a 'ls' command to the server
 * parameters:
 * o	fd : socket descriptor
 * o	p : pointer on a packet variable
 * o	s : size of 'p'
 */
void send_ls_ack( int fd , packet *p , size_t s )
{
	p->header = CTRL_PACKET ;
	p->type_packet.control.ordre = ACK_LS ;
	send( fd , p , s , 0 ) ;
}

/*
 * ls:
 * This function is the 'ls' for the server part.
 * parameters:
 * o	fd : socket descriptor
 */
void ls( int fd , struct sockaddr_in *s)
{
	packet		p ;
	int 		i , nb , ls_ok , retval , lg ;
	bloc_ls 	*tab = NULL;


	lg = sizeof( struct sockaddr_in );
	ls_ok = 1 ;
	tab = pre_ls( &nb );
	p.type_packet.control.ordre = ACK_LS;
	p.type_packet.control.gen_num = nb ;
	if( resend( low_fdsock , &p , sizeof(packet) , ACK_LS , (struct sockaddr *)s , sizeof( struct sockaddr) ) < 0 ) {
		free( tab_packet );
		close( low_fdsock );
		kill_lowserver( );
	}
				

	while( ls_ok != 0 ) {
		p.type_packet.control.ordre = DATA_LS ;
		for( i=0 ; i<nb ; i++ ) {
			if( tab != NULL ){ 
				if( tab[i].ok == -1 ) {
					p.type_packet.control.gen_num = i  ;
					bzero( p.type_packet.control.liste , NB_PACKET_TO_SEND );
					memcpy( p.type_packet.control.liste , tab[i].ligne_ls , NB_PACKET_TO_SEND );	
					sendto( low_fdsock , &p , sizeof(packet) , 0 , (struct sockaddr *)s , sizeof( struct sockaddr) );
				}
			} else {
				fprintf(stderr,"lowserver>> ls> ls array pointer non allocated !\n");
				exit(1);
			}
		} /* for */

		p.type_packet.control.ordre = END_LS ;
		p.type_packet.control.gen_num = nb ;
		if( resend( low_fdsock , &p , sizeof(packet) , ACK_END_LS , (struct sockaddr *)s , sizeof( struct sockaddr) ) < 0 ) {
			free( tab_packet );
			close( low_fdsock );
			kill_lowserver( );
		}

		retval = my_read( fdsock , &p , sizeof(packet) , (struct sockaddr *)s , &lg ); 
		if( retval == -1 ) {
			printf("lowserver> ls> my_read> waiting for ls ended...too long\n");
			kill_lowserver( );
		} else {
			switch(  p.type_packet.control.ordre ) {
				case LS_OK : {
					ls_ok = 0 ;
				}
				break; 
				case LS_NOT_OK: {
					/* this a LS_NOT_OK message */
					update_ls_block( tab , p.type_packet.control.liste , nb );
					ls_ok = 1 ;
				}
				break;
				default:
					printf("J'AI EU QQ PROBLEME DE COMPREHENSION !!\n");
			}
		}
	}/*while*/

	/* libration de la mmoire alloue */
	free( tab );
}

/*
 * pre_ls:
 * This function prepare ls packet.
 * parameters:
 * o	tab : arrays of caracters (packet)
 * return:
 * o	number of packet 
 * o	pointer on packets array 
 */
bloc_ls* pre_ls( int *nb )
{
	int 			i,j,k,taille_tab_pointeur,taille_champ,size_temp_p;
	DIR			*rep ;
	struct dirent		*p_rep ;
	char			buffer[NB_PACKET_TO_SEND];
	bloc_ls			*tmp  = NULL ;
	static char			*r = NULL ;


	i=0;
	
	taille_tab_pointeur= (i+1) * sizeof(bloc_ls);

	/* array of char which describes the directory */
	rep = opendir( "." );
	if( rep == NULL ) {
		perror("lowserver>> pre_ls> opendir");
	}

	taille_champ = 0 ;
	r = NULL ;
	while( (p_rep=readdir( rep )) != NULL ) {
		/* +2 : pour '\0' et ':' */
		taille_champ += strlen(p_rep->d_name) + 2 ;
		r = (char *) realloc( (void *)r , taille_champ*sizeof(char) );
		if( r == NULL ) {
			perror("realloc");
			exit(1);
		}

		r[ taille_champ-1 ] = '\0' ;
		r = strcat( r , p_rep->d_name );
		r = strcat( r , ":" );
	}
	closedir (rep ) ;

	/* calculate of nb strings and we give to 'nb' function parameter */
	if( ((int) strlen(r) % NB_PACKET_TO_SEND) == 0) {
		*nb = (int)strlen(r)/(NB_PACKET_TO_SEND-1)  ;
	} else {
		*nb = (int)strlen(r)/(NB_PACKET_TO_SEND-1) + 1 ;  
	}
	
	j = 0 ;
	tmp = NULL ;
	/* now we cut the string in several strings */
	for( i=0 ; i<strlen(r) ;i++ ){
		k = (int) i / (NB_PACKET_TO_SEND-1) ;	

		if( ( i % (NB_PACKET_TO_SEND-1) ) == 0 ) {

			if( k > 1 ) 
				tmp[ k-1 ].ligne_ls[j] = '\0' ;	

			tmp = (bloc_ls *) realloc( tmp , (k+1)*sizeof(bloc_ls) );
			if( tmp == NULL ) {
				perror("realloc");
				exit(2);
			}
			j = 0 ;
		}
		tmp[ k ].ligne_ls[j] = r[i];
		tmp[ k ].ok = -1 ;
		j++ ;
	}

	/* ne pas oublier de librer r */	
	free( r );

return tmp ;
}
